/*
 * Created on May 3, 2005
 *
 * Last Modified on July 13, 2006
 * Version: 0.4
 */
/**
 * @author zAlbee
 */
import java.io.*;
import java.util.*;
import java.text.*;

public class immerge {

	static final String NL = System.getProperty("line.separator"); // newline (CR/CRLF/LF)

	// The Byte-Order Mark (BOM), when occuring at the beginning of a file, indicates a Unicode file 
	// It is the single Unicode character U+FEFF (ZERO WIDTH NO-BREAK SPACE)
	// Generally, it is a convention of Win32 programs to place the BOM, but not POSIX systems.
	static final char UNICODEBOM = 0xFEFF; // BOM as Unicode character
	static final byte[] UTF8BOM = {(byte)0xEF, (byte)0xBB, (byte)0xBF}; // BOM encoded in UTF-8

	static int tz_adjust = -16;

	/***
	 * General Log class, associated with a BufferedReader
	 *
	 * Subclasses need to implement:
	 *
	 *   readSess() - reads the next session into a buffer
	 *              - puts that session's date into the sessDate property.
	 *              - returns True if a session was read, False if the log was empty (no session read)
	 *
	 *   getBuffer() - should return all the text within that session.
	 */
	static abstract class Log implements Comparable {
		static final String sessStart  = "Session Start (";
		static final String sessStart2 = "sion Start (";
		static final String sessClose  = "Session Close (";
		static SimpleDateFormat sdfTrillian = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
		static SimpleDateFormat sdfICQ = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss", Locale.US);
		static SimpleDateFormat sdfTTime = new SimpleDateFormat("[hh:mm a]", Locale.US);
		static SimpleDateFormat sdfTTime24 = new SimpleDateFormat("[HH:mm]", Locale.US);

		String filename;
		String pathname;
		BufferedReader reader;
		Date sessDate;
		String buffer;
		boolean empty = false;

		abstract boolean readSess() throws Exception;
		abstract String getBuffer();

		/** Log() contructor
		 */
		Log(String fname, String pname, File f) throws IOException {
			/*
			sdfTrillian.setTimeZone(TimeZone.getTimeZone("UTC"));
			sdfICQ.setTimeZone(TimeZone.getTimeZone("UTC"));
			sdfTTime.setTimeZone(TimeZone.getTimeZone("UTC"));
			*/
			filename = fname;
			pathname = pname;
			
			reader = new BufferedReader(new FileReader(f));
			
			reader.mark(16);
			if (reader.read() == 0xEF && reader.read() == 0xBB && reader.read() == 0xBF) { 
				// Detected UTF-8 header: ﻿
				// Use UTF-8 reader, but we still need to ignore the first 3 bytes (1 character).
				
				InputStreamReader in = new InputStreamReader(new FileInputStream(f), "utf-8");
				reader = new BufferedReader(in);
				
				reader.mark(16);
				if (reader.read() == 0xFEFF) {
				}
				else {
					// As long as Java's UTF-8 reader works, this should never happen
					// Maybe if the system is byte-swapped.
					System.out.println("Warning: UTF-8 BOM was detected, but was not treated as such. Send a bug report to zalbee@gmail.com");
					reader.reset();
				}
				
			}
			else reader.reset();
		}
		
		public int compareTo(Object o2) {
			Log log2 = (Log)o2;
			return sessDate.compareTo(log2.sessDate);
		}

	}

	/***
	 * TrLog class, subclass of Log
	 * Represents a Trillian logfile.
	 */
	static class TrLog extends Log {
		boolean sessOpen = false;
		boolean startOfFile = true;

		/** TrLog() contructor
		 */
		TrLog(String fname, String pname, File f) throws IOException {
			super(fname, pname, f);
		}

		/** TrLog.readSess()
		 */
		boolean readSess() throws Exception {
			sessOpen = false;
			String s = reader.readLine();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if (    ( s.startsWith( sessStart ) || (startOfFile && s.startsWith(sessStart2)) )
					 && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1
					) {
					int dateStartIdx = nameEndIdx + 3;

					sessDate = sdfTrillian.parse(s.substring(dateStartIdx));

					if (s.startsWith("sion")) {
						s = "Ses" + s;
					}
					sessOpen = true;
					buffer = s + NL;
				}
				else if (s.trim().length()>0) {
					// Text exists outside a session. Write it anyway to keep fidelity of logs.
					// Give it a "dummy" session, with the same date as the last one, so that it's written right away.
					// Since there will be no session close marker, Read until the next session starts, then rewind the reader.
					
					/*
					Calculation of max rewind for   BufferedReader.mark(int readAheadLimit)
					
					"Session Start (email@hotmail.com:MY LONG MSN DISPLAY NAME): Sun Jul 25 00:54:39 2004"
					"Session Start (".length() == 15;
					"): Sun Jul 25 00:54:39 2004".length() == 27
					Email address max == 64 + '@' + 255
						  		  	  == 320
					MSN name max == 129;
					----
					Total == 15 + 320 + ":" + 129 + 27
						  == 492
					*/
					
					buffer = s + NL;
					reader.mark(512);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.startsWith( sessStart ) && s.indexOf("):", s.length()-28) > -1 ) {
							reader.reset();
							break;
						}
						buffer += s + NL;
						reader.mark(512);
						s = reader.readLine();
					}
					
					startOfFile = false;
					return true;
					
					//System.out.println("Warning: Skipped line = " + s);
				}
				
				startOfFile = false;
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				buffer += s + NL;

				if ( s.startsWith( sessClose ) && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1 ) {
			  		sessOpen = false;
			  		buffer += NL + NL;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			return true;

		}

		String getBuffer() {
			return buffer;
		}
	} //end class TrLog


	/***
	 * HtmlLog class, subclass of Log
	 * Represents an ICQ log file (ICQ History Reader HTML format)
	 */
	static class HtmlLog extends Log {
		String yourNick;
		String nickName;
		int yourUIN;
		String theirUIN;
		String sendColour;
		String recvColour;
		static final String genStart = "<font size=3 color=";
		static final String genStart2 = ">";
		static final String genEnd = "</font>";
		static final int START_LEN = genStart.length() + 8;
		String sendStart;
		String recvStart;
		
		String msgBuffer;
		Date msgDate;
		boolean msgIsSend;

		String buffer2;
		Date buffer2Date = null;
		Date sessEndDate;

		/** HtmlLog() contructor
		 */
		HtmlLog(String fname, String pname, File f) throws IOException {
			super(fname, pname, f);
		}
		
		void init(int yourUIN, String theirUIN, String yourNick) throws IOException{
			this.yourUIN = yourUIN;
			this.theirUIN = theirUIN;
			this.yourNick = yourNick;
			
			String s = reader.readLine();

			while (s != null) {
				if (s.matches("<td [^>]*>Nickname:</td>")) {
					s=reader.readLine();
					nickName = s.substring( s.indexOf('>') +1, s.lastIndexOf("</td>") );
					break;
				}
				else if (s.trim().toLowerCase().equals("no info for this user")) {
					System.out.println("Info: No nickname in " + filename);
					nickName = "" + theirUIN;
					s = reader.readLine();
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Reached end-of-file while looking for info in " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Sent</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					sendColour = s.substring( colIdx, colIdx + 7);
					sendStart = genStart + sendColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read sent colour from " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Recieved</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					recvColour = s.substring( colIdx, colIdx + 7);
					recvStart = genStart + recvColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read receive colour from " + filename);
				empty = true;
				return;
			}

		}

		/** HtmlLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 */
		boolean readMsg() throws Exception {
			boolean msgOpen = false;
			//boolean isSend;
			msgBuffer = "";

			String s=reader.readLine();
			msgDate = null;
			//dpoint[0] = null;

			// grab date

			//while (s != null && dpoint[0]==null) {
			while (s != null && msgDate==null) {
				if ( s.startsWith(genStart) ) {
					int dateStartIdx = s.indexOf('(', START_LEN) + 1;
					int dateEndIdx = s.lastIndexOf(')');
					msgDate = sdfICQ.parse(s.substring(dateStartIdx, dateEndIdx));
					//dpoint[0].setTime( dpoint[0].getTime() - 16*60*60*1000 ); // has sync problems when dates are near daylight savings switch
					msgDate.setHours( msgDate.getHours() + tz_adjust); // -16 hours seems to make the time correct? no idea how ICQHR calculates timezone
				}
				s = reader.readLine();
			}
			
			// grab colour (to determine sender) and message

			while (s != null && !msgOpen) {
				if ( s.startsWith(genStart)) {
					msgOpen = true;
					msgIsSend = s.startsWith(sendStart);

					//msgBuffer = sdfTTime.format(dpoint[0]) + " ";
					//if (msgIsSend) msgBuffer += yourNick + ": ";
					//else msgBuffer += nickName + ": ";

					if (s.trim().endsWith(genEnd)) {
						msgBuffer += s.substring( START_LEN, s.lastIndexOf(genEnd)) + NL; msgOpen = false; return true;}//break;}
					else if (s.trim().endsWith("<BR>")) msgBuffer += s.substring( START_LEN, s.lastIndexOf("<BR>") ) + NL;
					else msgBuffer += s.substring( START_LEN ) + NL;
				}

				s=reader.readLine();
			}

			// if message consists of more than one line

			while (s != null && msgOpen) {
				if (s.trim().endsWith(genEnd)) {
					msgBuffer += s.substring( 0, s.lastIndexOf(genEnd)) + NL; msgOpen = false; return true;} //break;}
				else if (s.trim().endsWith("<BR>")) msgBuffer += s.substring( 0, s.lastIndexOf("<BR>") ) + NL;
				else msgBuffer += s + NL;

				s=reader.readLine();
			}
			
			// if we're here, message didn't terminate, OR there was no message in the first palce
			
			if (msgOpen) return true;
			else return false;

			//return msg;
		}
		
		/** HtmlLog.getMsg()
		 */
		String getMsg() {
			String prefix = sdfTTime.format(msgDate) + " ";
			if (msgIsSend) prefix += yourNick + ": ";
			else prefix += nickName + ": ";
			
			return prefix + msgBuffer;
		}

		/** HtmlLog.readSess()
		 */
		boolean readSess() throws Exception {
			//Date sDate = buffer2Date;
			//Date eDate;
			sessDate = buffer2Date;
			String msg = null;
			//Date[] dp = new Date[] {null};

			if (sessDate == null) {
				//m = readMsg(dp);
				boolean read = readMsg();
				if (!read) {empty=true; reader.close(); return false;}
				msg = getMsg();
				sessDate = msgDate;
				//if (m.length()==0) {empty=true; reader.close(); return false;}
				//sDate = dp[0];
			}
			else {
				msg = buffer2;
				buffer2 = null;
				buffer2Date = null;
			}

			//sessDate = eDate = sDate;
			sessEndDate = sessDate;
			//buffer = sessStart + (yourUIN>0 ? ""+yourUIN:"ICQ") + ":" + nickName + "): " + sdfTrillian.format(sDate) + NL;
			buffer = unHTMLEntity(msg);

			//m = readMsg(dp);
			boolean hasMsg = readMsg();

			//while (m.length() > 0) {
			while (hasMsg) {
				msg = getMsg();
				//Date date = dp[0];
				//Date date = msgDate;
				if (msgDate.getTime() < sessDate.getTime() + 24*60*60*1000) {
					buffer += unHTMLEntity(msg);
					sessEndDate = msgDate;
				}
				else {
					buffer2 = msg;
					buffer2Date = msgDate;
					break;
				}

				//m = readMsg(dp);
				hasMsg = readMsg();
			}

			//buffer += sessClose + nickName + "): " + sdfTrillian.format(sessEndDate) + NL+NL+NL;

			return true;
		}

		String getBuffer() {
			String prefix = sessStart + (yourUIN>0 ? ""+yourUIN:"ICQ") + ":" + nickName + "): " + sdfTrillian.format(sessDate) + NL;
			String suffix = sessClose + nickName + "): " + sdfTrillian.format(sessEndDate) + NL+NL+NL;

			return prefix + buffer + suffix;
		}

	} //end class HtmlLog


	/***
	 * ICQtoXML class, subclass of HtmlLog
	 * Represents the XML version of the ICQ History Reader file
	 */
	static class ICQtoXML extends HtmlLog {
		/** ICQtoXML() contructor
		 */
		ICQtoXML(String fname, String pname, File f) throws IOException {
			super(fname, pname, f);
		}
		
		String getMsg() {
			//<message type="incoming_privateMessage" time="1132707630" medium="MSN" to="zalbee2%40hotmail%2Ecom" from="catty_kitty_leung%40hotmail%2Ecom" from_display="%28%40%29%20SuPeR%20cAt%20%2D%20Jay%20Chow%20is%20amazing" text="that%20was%20a%20FAST%20response"/>
			//<message type="outgoing_privateMessage" time="1133042238" medium="ICQ" to="16225800" from="18196095" from_display="zAlbee" text="hm%2E"/>
			String prefix = "<message type=\"" + (msgIsSend ? "outgoing_privateMessage" : "incoming_privateMessage")
					+ "\" time=\"" + msgDate.getTime()/1000 + "\" medium=\"ICQ\" to=\""
					+ (msgIsSend ? theirUIN : ""+yourUIN) + "\" from=\"" + (msgIsSend ? ""+yourUIN : theirUIN)
					+ "\" from_display=\"" + (msgIsSend ? yourNick : nickName) + "\" text=\"";
			
			String suffix = "\"/>" + NL;
			
			String tmpMsg = msgBuffer;
			
			if (tmpMsg.endsWith("\r\n")) {
				tmpMsg = tmpMsg.substring(0, tmpMsg.length() - 2);
			}
			else if (tmpMsg.endsWith("\n") || tmpMsg.endsWith("\r")) {
				tmpMsg = tmpMsg.substring(0, tmpMsg.length() - 1);
			}

			return prefix + numericEntity(tmpMsg) + suffix;
		}

		String getBuffer() {
			//<session type="start" time="1133041383" medium="ICQ" to="16225800" from="18196095"/>
			String prefix = "<session type=\"start\" time=\"" + sessDate.getTime()/1000 +  "\" medium=\"ICQ\" to=\"" 
					+ theirUIN +"\" from=\""+ yourUIN + "\"/>" + NL;

			//<session type="stop" time="1133060545" medium="ICQ" to="16225800" from="18196095"/>
			String suffix = "<session type=\"stop\" time=\"" + sessEndDate.getTime()/1000 +  "\" medium=\"ICQ\" to=\"" 
					+ theirUIN +"\" from=\""+ yourUIN + "\"/>" + NL;
					
			return prefix + buffer + suffix;
		}
	}


	/***
	 * QueryXML class, subclass of Log
	 * Represents a Trillian XML log file (in the "query" folder)
	 */
	static class QueryXML extends Log {
		boolean sessOpen = false;
		
		/** QueryXML() contructor
		 */
		QueryXML(String fname, String pname, File f) throws IOException {
			super(fname, pname, f);
		}

		boolean readSess() throws Exception {
			sessOpen = false;
			String s = reader.readLine();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
					
					int dateStartIdx = s.indexOf("time=\"") + 6;
					int dateEndIdx = s.indexOf("\"", dateStartIdx);
					
					sessDate = new Date( Long.parseLong( s.substring(dateStartIdx, dateEndIdx) ) );
					
					sessOpen = true;
					buffer = s + NL;
				}
				else if (s.trim().length()>0) {
					//System.out.println("Warning: skipping line: " + s);
					buffer = s + NL;
					reader.mark(1024);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
							reader.reset();
							break;
						}
						buffer += s + NL;
						reader.mark(1024);
						s = reader.readLine();
					}
					return true;
				}
				
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				buffer += s + NL;

				if ( s.toLowerCase().startsWith( "<session type=\"stop\"" ) ) {
			  		sessOpen = false;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			return true;
			
		}
		
		String getBuffer() {
			return buffer;
		}
		
	} //end class QueryXML


	/**
	 * unHTMLEntity ( String )
	 *
	 * Returns a new string with various HTML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 */
	static String unHTMLEntity(String s) {
		return s.replaceAll("&amp;","&").replaceAll("&nbsp;"," ").replaceAll("&lt;","<").replaceAll("&gt;",">");
	}
	
	/**
	 * numericEntity ( String )
	 *
	 * Returns a new string with all characters not in the set [0-9A-Za-z_] replaced with %hh
	 * where hh is the character's byte value in hex.
	 *
	 * Generally hh is two digits, as most chars are 1-byte. However Java String natively supports 
	 * 2-byte Unicode chars, in which case you may get %hhh or %hhhh.
	 *
	 * This function is equivalent to rawurlencode() in PHP for all 1-byte characters.
	 */
	 
	static String numericEntity(String s) {
		String out = "";
		for (int i=0; i<s.length(); i++) {
			char c = s.charAt(i);
			if ( ('0'<=c && c<='9') || ('a'<=c && c<='z') || ('A'<=c && c<='Z') || c=='_' ) {
				out += c;
			}
			else {
				String hex = Integer.toHexString(c).toUpperCase();
				String hh = hex;
				if (hh.length() < 2) hh = "0"+hh;
				out += "%"+hh;
			}
		}
		return out;
	}

	/**
	 * readIntFrom ( BufferedReader )
	 *
	 * Reads an integer from the stream, returns it.
	 * Will reprompt and retry the read until an integer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static int readIntFrom(BufferedReader reader) throws IOException{
		String utxt = reader.readLine();
		try {
			int i = Integer.parseInt(utxt.trim());
			return i;
		} catch (Exception e) {
			System.out.print("You must enter a number: ");
			return readIntFrom(reader);
		}
	}
	
	/**
	 * readYesNoFrom ( BufferedReader )
	 *
	 * Reads the stream for Yes or No, returning True or False, respectively.
	 * Will reprompt and retry the read until an answer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static boolean readYesNoFrom(BufferedReader reader) throws Exception{
		boolean yesno = false;
		String utxt = reader.readLine().trim() + " ";
		char first = utxt.charAt(0);
		if (first == 'Y' || first == 'y') return true;
		else if (first == 'N' || first == 'n') return false;
		else {
			System.out.print("Yes or No? ");
			return readYesNoFrom(reader);
		}
	}

	/***
	 *  main ( String[] )
	 *
	 *  Asks the user for some information.
	 *  Then executes the merging of Trillian logs.
	 ***/
	public static void main(String[] args) throws Exception {
		int numTrLog, numHtml, yourUIN = 0;
		String yourNick = null;

		String[] mediums = {"AIM", "ICQ", "IRC", "MSN", "YAHOO"};
		String outPath = "output";
		String[] outPath2 = new String[5];

		File outDir = new File(outPath);
		if (!outDir.isDirectory() && !outDir.mkdir()) {
			System.out.println("Couldn't create '" + outPath + "' directory");
			System.out.println("Failed.");
			return;
		}

		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		System.out.print("How many folders (logs1 .. logsN) of Trillian-style <username>.log: ");
		numTrLog = readIntFrom(in);
		System.out.print("How many folders (icqh1 .. icqhN) of IcqHR output <UIN>.html (UIN per file): ");
		numHtml = readIntFrom(in);

		if (numHtml > 0) {
			System.out.print("Your ICQ UIN/number (required): ");
			yourUIN = readIntFrom(in);
			System.out.print("Your ICQ Nickname (as you want it to appear): ");
			yourNick = in.readLine();
			System.out.print("Correct ICQHR timezone by how many hours? (-16 = default, 0 = no change): ");
			tz_adjust = readIntFrom(in);
			System.out.print("Display time in 24-hr format? (yes = 24-hr, no = AM/PM): ");
			boolean use24h = readYesNoFrom(in);
			if (use24h) {
				Log.sdfTTime = Log.sdfTTime24;
			}
		}

		System.out.print("Should the logs be output to a \"Query\" folder? (Yes/No): ");
		boolean useQuery = readYesNoFrom(in);
		
		System.out.println();

		for (int j=0; j<mediums.length; j++) {
			System.out.println("== " + mediums[j] + " ==");

			// link the username to a Vector of the logs
			Hashtable userLogs = new Hashtable();

			// link the username to a Vector of the XML files
			Hashtable userXMLs = new Hashtable();

			// Find files - Trillian logs

			for (int i=0; i<numTrLog; i++) {
				String logsPath = "logs" + (i+1) + File.separator + mediums[j]; // in folder
				String queryPath = logsPath + File.separator + "Query";         // in Query folder
				
				File dir = new File(logsPath);
				File dirQ = new File(queryPath);
				if (!dir.isDirectory()) continue;

				File[][] files = new File[2][];
				files[0] = dir.listFiles();

				if (dirQ.isDirectory())
					files[1] = dirQ.listFiles();
				else
					files[1] = new File[0];
					
				for (int qd = 0; qd <= 1; qd++) {
					for (int k=0; k<files[qd].length; k++) {
						File f = files[qd][k];
						String filename = f.getName();

						// Trillian .log
						if (f.isFile() && filename.endsWith(".log")) {
							TrLog lg = new TrLog( filename, f.getPath(), f );
							String user = filename.substring(0, filename.length()-4);
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}
						}

						// Trillian .xml
						if (qd==1 && f.isFile() && filename.endsWith(".xml")) {
							
							if (filename.endsWith("-assets.xml")) {
								// assets, don't process - Trillian finds automatically
							}
							else {
								QueryXML xml = new QueryXML( filename, f.getPath(), f );
								String user = filename.substring(0, filename.length()-4);
								if (userXMLs.containsKey(user))
									((Vector)userXMLs.get(user)).add(xml);
								else {
									Vector XMLs = new Vector();
									XMLs.add(xml);
									userXMLs.put(user, XMLs);
								}
							}
							
						}
						
					} // k
				} // qd
				
			} // i (numTrLog)

		
			// Find files - ICQ History Reader 1.8 .html output
			if (j==1) { // ICQ

				for (int i=0; i<numHtml; i++) {
					File dir = new File("icqh" + (i+1));
					if (!dir.isDirectory()) continue;

					File[] files;
					files = dir.listFiles();
					for (int k=0; k<files.length; k++) {
						String filename = files[k].getName();

						if (files[k].isFile() && filename.endsWith(".html")) {
							String user = filename.substring(0, filename.length()-5);
							HtmlLog lg = new HtmlLog( filename, files[k].getPath(), files[k] );
							//lg.yourNick = yourNick;
							//lg.yourUIN = yourUIN;
							//lg.theirUIN = user;
							lg.init(yourUIN, user, yourNick);
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}

							ICQtoXML xml = new ICQtoXML( filename, files[k].getPath(), files[k] );
							xml.init(yourUIN, user, yourNick);
							if (userXMLs.containsKey(user))
								((Vector)userXMLs.get(user)).add(xml);
							else {
								Vector XMLs = new Vector();
								XMLs.add(xml);
								userXMLs.put(user, XMLs);
							}

						}
					}
				}
			}
			

			/** ---- merge Logs ---- **/

			// if logs exist, make output dir
			if (userLogs.size() > 0) {
				outPath2[j] = outPath + File.separator + mediums[j] + (useQuery? File.separator+"Query" : "");
				
				File outDir2 = new File(outPath2[j]);
				if (!outDir2.isDirectory() && !outDir2.mkdirs()) {
					System.out.println("Couldn't create directory '" + outPath2[j] + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}
			}

			// Iterate through the logs.
			int users = 0;
			Enumeration userEnum = userLogs.keys();
			Vector userNames = new Vector();

			while (userEnum.hasMoreElements()) {
				users++;
				userNames.add((String)userEnum.nextElement());
			}
			Collections.sort(userNames);

			//while (userNames.hasMoreElements()) {
			for (int u=0; u<users; u++) {
				int sessions = 0;
				int dupes = 0;
				int numFiles;

				String user = (String)userNames.elementAt(u);
				Vector logs = (Vector)userLogs.get(user);

				// Initialize; grab one session from each log, and record in log buffer
				for (int i=0; i<logs.size(); i++) {
					Log lg = (Log)logs.get(i);
					lg.sessDate = new Date(i); // initialize date, in case the readSess() func doesn't find a session date.
					if (!lg.readSess()) logs.remove(i--);
				}
				Collections.sort(logs);
				numFiles = logs.size();

				File f = new File(outPath2[j] + File.separator + user + ".log");
				if (!f.exists()) f.createNewFile();

				try {
					OutputStreamWriter fw = new OutputStreamWriter(new FileOutputStream(f), "utf8");
					fw.write(UNICODEBOM);
					
					//fw.write(0xFEFF);

					while (logs.size()>0) {
						sessions++;
						Log lg = (Log)logs.get(0);
						fw.write(lg.getBuffer());

						for (int i=1; i<logs.size(); i++) {
							Log olog = (Log)logs.get(i);
							if (!lg.sessDate.equals(olog.sessDate)) {
								//System.out.println("No Dup - " + lg.sessDate + " vs " + olog.sessDate);
								break;
							}

							dupes++;
							//System.out.println("-- Duplicate in " + lg.filename + "(" + lg.pathname + " & " + olog.pathname + ") on " + lg.sessDate);
							if (!olog.readSess()) logs.remove(i--);
						}

						if (!lg.readSess()) logs.remove(0);
						else Collections.sort(logs);
					}

					fw.close();

					System.out.println(user + ": merged " + numFiles + " files, " + sessions + " sessions, " + dupes + " duplicates");
				}
				catch (IOException e) {
					e.printStackTrace();
				}

			} //end for users

			System.out.println(users + " Log files written.\n");


			/** ---- merge XML ---- **/

			// if XML files exist, make output dir
			if (userXMLs.size() > 0) {
				outPath2[j] = outPath + File.separator + mediums[j] + File.separator + "Query";
				
				File outDir2 = new File(outPath2[j]);
				if (!outDir2.isDirectory() && !outDir2.mkdirs()) {
					System.out.println("Couldn't create directory '" + outPath2[j] + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}
			}

			// Iterate through the XML files.
			users = 0;
			userEnum = userXMLs.keys();
			userNames = new Vector();

			while (userEnum.hasMoreElements()) {
				users++;
				userNames.add((String)userEnum.nextElement());
			}
			Collections.sort(userNames);

			for (int u=0; u<users; u++) {
				int sessions = 0;
				int dupes = 0;
				int numFiles;
				
				int numMinus = 0;
				int numMinusDupes = 0;
				Date minusDate = new Date(-1);

				String user = (String)userNames.elementAt(u);
				Vector XMLs = (Vector)userXMLs.get(user);

				// Initialize; grab one session from each log, and record in log buffer
				for (int i=0; i<XMLs.size(); i++) {
					Log xml = (Log)XMLs.get(i);
					xml.sessDate = new Date( i - XMLs.size() - 1); // initialize date, in case the readSess() func doesn't find a session date.
					if (!xml.readSess()) XMLs.remove(i--);
				}
				Collections.sort(XMLs);
				numFiles = XMLs.size();

				File f = new File(outPath2[j] + File.separator + user + ".xml");
				if (!f.exists()) f.createNewFile();

				try {
					FileWriter fw = new FileWriter(f);
					//fw.write(UTF8BOM); // Trillian doesn't seem to UTF-8 the XML files


					while (XMLs.size()>0) {
						sessions++;
						Log xml = (Log)XMLs.get(0);
						fw.write(xml.getBuffer());

						if (xml.sessDate.equals(minusDate))
							numMinus++;

						for (int i=1; i<XMLs.size(); i++) {
							Log oxml = (Log)XMLs.get(i);
							if (!xml.sessDate.equals(oxml.sessDate) ) {
								//System.out.println("No Dup - " + xml.sessDate + " vs " + oxml.sessDate);
								break;
							}

							if (xml.sessDate.equals(minusDate)) {
								// if we find two sessions with -1, we can't be sure that they are equal, but we will treat them that way
								numMinusDupes++;
							}
							dupes++;
							//System.out.println("-- Duplicate in " + xml.filename + "(" + xml.pathname + " & " + oxml.pathname + ") on " + xml.sessDate);
							if (!oxml.readSess()) XMLs.remove(i--);
						}

						if (!xml.readSess()) XMLs.remove(0);
						else Collections.sort(XMLs);
					}

					fw.close();

					System.out.println(user + ": merged " + numFiles + " XML files, " + sessions + " sessions, " + dupes + " duplicates");
					if (numMinus > 0)
						System.out.println(" * Warning: kept " + numMinus + " sessions with a -1 date, an additional " + numMinusDupes + " may have been incorrectly dropped as duplicates");
				}
				catch (IOException e) {
					e.printStackTrace();
				}

			} //end while enumeration of users

			System.out.println(users + " XML files written.\n");

		} //end for mediums

		in.close();

		System.out.println("Finished.");

	} // end main

}
